"""Status Update."""
from __future__ import annotations

from datetime import datetime


class StatusUpdate:
    """Status Update class."""

    def __init__(
        self,
        status: str,
        last_update: datetime | None,
        last_update_successful: datetime | None,
        last_timestamp: datetime | None,
        total: int,
        created: int,
        updated: int,
        removed: int,
    ):
        """Initialise this status update."""
        self._status = status
        self._last_update = last_update
        self._last_update_successful = last_update_successful
        self._last_timestamp = last_timestamp
        self._total = total
        self._created = created
        self._updated = updated
        self._removed = removed

    def __repr__(self):
        """Return string representation of this entry."""
        return "<{}({}@{})>".format(
            self.__class__.__name__, self.status, self.last_update
        )

    @property
    def status(self) -> str:
        """Return the status."""
        return self._status

    @property
    def last_update(self) -> datetime | None:
        """Return the time when the feed was last updated."""
        return self._last_update

    @property
    def last_update_successful(self) -> datetime | None:
        """Return the time when the feed was last updated successfully."""
        return self._last_update_successful

    @property
    def last_timestamp(self) -> datetime | None:
        """Return the timestamp the latest entry in the feed."""
        return self._last_timestamp

    @property
    def total(self) -> int:
        """Return the total number of managed entries."""
        return self._total

    @property
    def created(self) -> int:
        """Return the number of created entries."""
        return self._created

    @property
    def updated(self) -> int:
        """Return the number of updated entries."""
        return self._updated

    @property
    def removed(self) -> int:
        """Return the number of removed entries."""
        return self._removed



"""XML Parser."""
from __future__ import annotations

import logging
from typing import Dict, Optional

import dateparser as dateparser
import xmltodict

from aio_quakeml_client.consts import (
    XML_TAG_CREATIONINFO,
    XML_TAG_CREATIONTIME,
    XML_TAG_DEPTH,
    XML_TAG_EVENT,
    XML_TAG_EVENTPARAMETERS,
    XML_TAG_LATITUDE,
    XML_TAG_LONGITUDE,
    XML_TAG_MAG,
    XML_TAG_MAGNITUDE,
    XML_TAG_ORIGIN,
    XML_TAG_Q_QUAKEML,
    XML_TAG_STATIONCOUNT,
    XML_TAG_TIME,
    XML_TAG_VALUE,
)
from aio_quakeml_client.xml_parser.event_parameters import EventParameters

_LOGGER = logging.getLogger(__name__)

DEFAULT_NAMESPACES = {
    "http://quakeml.org/xmlns/bed/1.2": None,
    "http://quakeml.org/xmlns/quakeml/1.2": "q",
}
KEYS_CHAINS_DATE = [
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_ORIGIN,
        XML_TAG_TIME,
        XML_TAG_VALUE,
    ],
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_CREATIONINFO,
        XML_TAG_CREATIONTIME,
    ],
]
KEYS_CHAINS_INT = [
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_ORIGIN,
        XML_TAG_DEPTH,
        XML_TAG_VALUE,
    ],
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_MAGNITUDE,
        XML_TAG_STATIONCOUNT,
    ],
]
KEY_CHAINS_FLOAT = [
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_ORIGIN,
        XML_TAG_LATITUDE,
        XML_TAG_VALUE,
    ],
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_ORIGIN,
        XML_TAG_LONGITUDE,
        XML_TAG_VALUE,
    ],
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_MAGNITUDE,
        XML_TAG_MAG,
        XML_TAG_VALUE,
    ],
]


class XmlParser:
    """Built-in XML parser."""

    def __init__(self, additional_namespaces: Dict = None):
        """Initialise the XML parser."""
        self._namespaces = DEFAULT_NAMESPACES
        if additional_namespaces:
            self._namespaces.update(additional_namespaces)

    @staticmethod
    def postprocessor(path, key, value):
        """Conduct type conversion for selected keys."""
        try:
            if XmlParser._is_path_in(path, KEY_CHAINS_FLOAT):
                return key, float(value)
            if XmlParser._is_path_in(path, KEYS_CHAINS_INT):
                return key, int(value)
            if XmlParser._is_path_in(path, KEYS_CHAINS_DATE):
                return key, dateparser.parse(
                    value,
                    settings={"TIMEZONE": "UTC", "RETURN_AS_TIMEZONE_AWARE": True},
                )
        except (ValueError, TypeError) as error:
            _LOGGER.warning("Unable to process (%s/%s): %s", key, value, error)
        return key, value

    @staticmethod
    def _is_path_in(path, chains):
        """Test if the path is in any of the provided chains."""
        if path and chains:
            new_path = []
            for element in path:
                new_path.append(element[0])
            for chain in chains:
                if chain == new_path:
                    return True
        return False

    def parse(self, xml) -> Optional[EventParameters]:
        """Parse the provided xml."""
        if xml:
            parsed_dict = xmltodict.parse(
                xml,
                process_namespaces=True,
                namespaces=self._namespaces,
                postprocessor=XmlParser.postprocessor,
            )
            if XML_TAG_Q_QUAKEML in parsed_dict:
                return XmlParser._create_feed_from_quakeml(parsed_dict)
        return None

    @staticmethod
    def _create_feed_from_quakeml(parsed_dict: Dict) -> Optional[EventParameters]:
        """Create feed from provided RSS data."""
        quakeml = parsed_dict.get(XML_TAG_Q_QUAKEML)
        if XML_TAG_EVENTPARAMETERS in quakeml:
            events = quakeml.get(XML_TAG_EVENTPARAMETERS)
            return EventParameters(events)
        else:
            _LOGGER.warning(
                "Invalid structure: Missing top level element %s", XML_TAG_Q_QUAKEML
            )
            return None



"""Feed Entry."""
from __future__ import annotations

import logging
import re
from abc import ABC, abstractmethod
from typing import Tuple

from haversine import haversine

from .consts import CUSTOM_ATTRIBUTE
from .xml_parser.creation_info import CreationInfo
from .xml_parser.event import Event
from .xml_parser.magnitude import Magnitude
from .xml_parser.origin import Origin

_LOGGER = logging.getLogger(__name__)


class FeedEntry(ABC):
    """Feed entry base class."""

    def __init__(self, home_coordinates: Tuple[float, float], quakeml_event: Event):
        """Initialise this feed entry."""
        self._home_coordinates = home_coordinates
        self._quakeml_event = quakeml_event

    def __repr__(self):
        """Return string representation of this entry."""
        return "<{}(id={})>".format(self.__class__.__name__, self.external_id)

    @property
    def coordinates(self) -> Tuple[float, float] | None:
        """Return the coordinates (latitude, longitude) of this entry."""
        if self.origin and self.origin.latitude and self.origin.longitude:
            return self.origin.latitude, self.origin.longitude
        return None

    @property
    def external_id(self) -> str | None:
        """Return the external id of this entry."""
        if self._quakeml_event:
            external_id = self._quakeml_event.public_id
            if not external_id:
                # Use geometry as ID as a fallback.
                external_id = hash(self.coordinates)
            return external_id
        return None

    def _search_in_external_id(self, regexp) -> str | None:
        """Find a sub-string in the entry's external id."""
        if self.external_id:
            match = re.search(regexp, self.external_id)
            if match:
                return match.group(CUSTOM_ATTRIBUTE)
        return None

    @property
    @abstractmethod
    def attribution(self) -> str | None:
        """Return the attribution of this entry."""
        return None

    @property
    def distance_to_home(self) -> float:
        """Return the distance in km of this entry to the home coordinates."""
        distance = float("inf")
        if self.coordinates:
            # Expecting coordinates in format: (latitude, longitude).
            return haversine(self.coordinates, self._home_coordinates)
        return distance

    @property
    def type(self) -> str | None:
        """Return entry's type."""
        if self._quakeml_event:
            return self._quakeml_event.type
        return None

    @property
    def description(self) -> str | None:
        """Return the description of this entry."""
        if self._quakeml_event and self._quakeml_event.description:
            if self._quakeml_event.description.type:
                return f"{self._quakeml_event.description.type.capitalize()}: {self._quakeml_event.description.text}"
            else:
                return self._quakeml_event.description.text
        return None

    @property
    def creation_info(self) -> CreationInfo | None:
        """Return creation info."""
        if self._quakeml_event:
            return self._quakeml_event.creation_info
        return None

    @property
    def magnitude(self) -> Magnitude | None:
        """Return magnitude."""
        if self._quakeml_event:
            return self._quakeml_event.magnitude
        return None

    @property
    def origin(self) -> Origin | None:
        """Return origin."""
        if self._quakeml_event:
            return self._quakeml_event.origin
        return None



"""Constants for feeds and feed entries."""

ATTR_ATTRIBUTION = "attribution"

CUSTOM_ATTRIBUTE = "custom_attribute"

DEFAULT_REQUEST_TIMEOUT = 10

UPDATE_OK = "OK"
UPDATE_OK_NO_DATA = "OK_NO_DATA"
UPDATE_ERROR = "ERROR"

XML_ATTR_PUBLICID = "@publicID"

XML_CDATA = "#text"

XML_TAG_Q_QUAKEML = "q:quakeml"
XML_TAG_AGENCYID = "agencyID"
XML_TAG_AUTHOR = "author"
XML_TAG_CREATIONINFO = "creationInfo"
XML_TAG_CREATIONTIME = "creationTime"
XML_TAG_DEPTH = "depth"
XML_TAG_DEPTHTYPE = "depthType"
XML_TAG_DESCRIPTION = "description"
XML_TAG_EVENT = "event"
XML_TAG_EVALUATIONMODE = "evaluationMode"
XML_TAG_EVALUATIONSTATUS = "evaluationStatus"
XML_TAG_EVENTPARAMETERS = "eventParameters"
XML_TAG_LATITUDE = "latitude"
XML_TAG_LONGITUDE = "longitude"
XML_TAG_MAG = "mag"
XML_TAG_MAGNITUDE = "magnitude"
XML_TAG_ORIGIN = "origin"
XML_TAG_STATIONCOUNT = "stationCount"
XML_TAG_TEXT = "text"
XML_TAG_TIME = "time"
XML_TAG_TYPE = "type"
XML_TAG_VALUE = "value"
